/**
 * eibnetip.c
 *
 * implementation of eibnet/ip
 *
 * @author: fritz
 */
 
#include "eibnetip.h"
#include "uartapi.h"
#include "config.h"
#include "ip.h"
#include "udp.h"
#include "stddef.h"
#include "convert.h"
#include "config.h"

//globals
EIBNETIP_CONNECTION eibcon[EIBNETIP_MAXCONNECTIONS]={0}; // saves all active connections
 
/**
 * implements eibnet/ip core
 * handles all packets to port EIBNETIP_PORT_NUMBER
 * e.g. SEARCH_REQUESTS, DESCRIPTION_REQUESTS, ...
 **/
void eibnetipserver(byte * rcvdata, word rcvdatalen) {
	CEMI_L_DATA_MESSAGE cemi;
	EIBFRAME eib;

	// get eibnet/ip header
	EIBNETIP_HEADER *head;

	#ifdef DEBUG_ENABLE_ALL
		int i=0;
		printfUART(UART_DEBUG,"received eibnetip frame: 0x");
		for (i=0;i<rcvdatalen;i++) {
			printfUART(UART_DEBUG,"%x",rcvdata[i]);
		}
		printfUART(UART_DEBUG,"\r");
	#endif


	head = (EIBNETIP_HEADER *) rcvdata; 
	
	// fix endian format
	head->servicetype=(rcvdata[2]<<8) | rcvdata[3];
	head->totalsize=(rcvdata[4]<<8)|rcvdata[5];
	
	
	// perform some checks to packet
	// @todo: send error codes to sender
	if (head->headersize!=HEADER_SIZE_10) {
		// wrong header size
		printfUART(UART_DEBUG,"received wrong header_size.\r");
		return;
	}
	if (head->version!=EIBNETIP_VERSION_10) {
		// wrong version
		printfUART(UART_DEBUG,"received unsupported eibnet/ip version.\r");
		return;
	}
	if (head->totalsize!=rcvdatalen) {
		// hmm... something went seriously wrong
		printfUART(UART_DEBUG,"received datasize does not match total size of eibnet/ip packet\r");
		return;	
	}
	

	// pass data to next layer according to service type
	switch (head->servicetype>>8) {
		case EIBNETIP_CORE:
			handleCore(&rcvdata[0], head->totalsize);
			break;

		case EIBNETIP_DEVMGMT:
			handleDeviceManagement(&rcvdata[0], head->totalsize);
			break;

		case EIBNETIP_TUNNELING:
			handleTunneling(&rcvdata[0], head->totalsize);
			break;			

		case EIBNETIP_ROUTING:
			// hack
			// smoke up eib bus and pass every ip packet to bus
			extractCemi(&cemi,&rcvdata[6]);
			
			initEIBframeCemi(&eib,&cemi);

			sendEIBframe(UART_KNX0,&eib);
			
			break;
	
		// @todo: handle other layers
		
		default:
			printfUART(UART_DEBUG,"received unsupported layer service type 0x%x%x.\r",head->servicetype>>8,head->servicetype&0xFF);
			break;
	}

}

/**
 * handles core service type
 **/
void handleCore(byte * rcvdata, word rcvdatalen) {

	// get eibnet/ip header
	EIBNETIP_HEADER *head;
	head = (EIBNETIP_HEADER *) rcvdata[0]; 


	// pass data to next layer according to service type
	switch (head->servicetype) {
		case SEARCH_REQUEST:
			handleSearchRequest(&rcvdata[head->headersize], head->totalsize-head->headersize);
			break;
		
		case DESCRIPTION_REQUEST:
			handleDescriptionRequest(&rcvdata[head->headersize], head->totalsize-head->headersize);
			break;
			
		case CONNECT_REQUEST:
			handleConnectRequest(&rcvdata[head->headersize], head->totalsize-head->headersize);
			break;

		case CONNECTIONSTATE_REQUEST:
			handleConnectionstateRequest(&rcvdata[head->headersize], head->totalsize-head->headersize);
			break;
			
		case DISCONNECT_REQUEST:
			handleDisconnectRequest(&rcvdata[head->headersize], head->totalsize-head->headersize);
			break;
		
		case DISCONNECT_RESPONSE:
			// ignore
			break;

		default:
			printfUART(UART_DEBUG,"received unsupported core service type 0x%x.\r",head->servicetype);
			break;
	}
}

/**
 * handle Device Management Service Type
 **/
void handleDeviceManagement(byte * rcvdata, word rcvdatalen) {
	EIBNETIP_COMMON_CONNECTION_HEADER ret_head;
	EIBNETIP_COMMON_CONNECTION_HEADER *head;
	EIBNETIP_HEADER *h;
	EIBNETIP_CONNECTION *c;

	h=(EIBNETIP_HEADER *)&rcvdata[0];
	head=(EIBNETIP_COMMON_CONNECTION_HEADER *)&rcvdata[sizeof(EIBNETIP_HEADER)-1];

	ret_head.typespecific=checkConnection(head);

	if (ret_head.typespecific!=E_NO_ERROR) 
		return; // some sort of error, do nothing


	// get connection
	c=&eibcon[head->channelid-1];

	ret_head.structlength=sizeof(EIBNETIP_COMMON_CONNECTION_HEADER);
	ret_head.channelid=c->channelid;
	ret_head.sequencecounter=c->sequencecounter_sent++;

	switch (h->servicetype) {
		case DEVICE_CONFIGURATION_REQUEST:
			eibnetipsend(&c->hpai, DEVICE_CONFIGURATION_ACK, (byte *)&ret_head,sizeof(EIBNETIP_COMMON_CONNECTION_HEADER));

			cEMI_Mgmt_Server((CEMI_MGMT_MESSAGE *)&rcvdata[sizeof(EIBNETIP_HEADER)-sizeof(EIBNETIP_COMMON_CONNECTION_HEADER)-1],rcvdatalen-sizeof(EIBNETIP_HEADER)-sizeof(EIBNETIP_COMMON_CONNECTION_HEADER));

			break;

		default:
			printfUART(UART_DEBUG,"received unsupported device management service type 0x%x.\r",h->servicetype);
			break;
	}
}

/**
 * handles tunneling service type
 **/
void handleTunneling(byte * rcvdata, word rcvdatalen) {
	EIBNETIP_COMMON_CONNECTION_HEADER ret_head;
	EIBNETIP_COMMON_CONNECTION_HEADER *head;
	EIBNETIP_HEADER *h;
	EIBNETIP_CONNECTION *c;
	CEMI_L_DATA_MESSAGE cemi;
	EIBFRAME eib;

	h=(EIBNETIP_HEADER *)&rcvdata[0];
	head=(EIBNETIP_COMMON_CONNECTION_HEADER *)&rcvdata[sizeof(EIBNETIP_HEADER)-1];

	ret_head.typespecific=checkConnection(head);

	if (ret_head.typespecific!=E_NO_ERROR) 
		return; // some sort of error, do nothing


	// get connection
	c=&eibcon[head->channelid-1];

	ret_head.structlength=sizeof(EIBNETIP_COMMON_CONNECTION_HEADER);
	ret_head.channelid=c->channelid;
	ret_head.sequencecounter=c->sequencecounter_sent++;

	switch (h->servicetype) {
		case TUNNELLING_REQUEST:
			eibnetipsend(&c->hpai, TUNNELLING_ACK, (byte *)&ret_head,sizeof(EIBNETIP_COMMON_CONNECTION_HEADER));

			// extract cemi frame
			extractCemi(&cemi,&rcvdata[sizeof(EIBNETIP_HEADER)+sizeof(EIBNETIP_COMMON_CONNECTION_HEADER)-1]);

			// check cemi frame
			if (cemi.saddr_l==0 && cemi.saddr_h==0) {
				// replace 0 srcaddress with address assigned to connection
				cemi.saddr_h=*c->knxaddress>>8;
				cemi.saddr_l=*c->knxaddress&0xFF;
			}
			
			// build eib frame
			initEIBframeCemi(&eib, &cemi);
			
			// send frame
			sendEIBframe(UART_KNX0,&eib);
			
			break;

		default:
			printfUART(UART_DEBUG,"received unsupported tunnelling service type 0x%x.\r",h->servicetype);
			break;
	}


}


/**
 * checks connection header
 * e.g sequence number, channelid, ...
 * increments sequence counter
 **/
byte checkConnection(EIBNETIP_COMMON_CONNECTION_HEADER *head) {

	if (eibcon[head->channelid-1].channelid==0) 
		return E_CONNECTION_ID;

	if ((head->sequencecounter!=eibcon[head->channelid-1].sequencecounter_rcv) ||
		(head->sequencecounter+1!=eibcon[head->channelid-1].sequencecounter_rcv)) 
		return E_SEQUENCE_NUMBER;

	eibcon[head->channelid-1].sequencecounter_rcv++;

//	if (head->connectiontype!=eibcon[head->channelid-1].connectiontype)
//		return E_DATA_CONNECTION;

	return E_NO_ERROR;
}

/**
 * adds EIBNET/IP Header and 
 * sends *senddata of length data_size (without eibnetip header size!!) with service type
 * service_type to endpoint receiver
 * according to protocol stated in HPAI either tcp or udp is used
 **/
void eibnetipsend(EIBNETIP_HPAI *receiver, word service_type, byte *senddata, word data_size) {
	IPHDATA iph;
	UDPHDATA udph;
	EIBNETIP_PACKET p;
	
	// prepare EIBNET/IP packet
	p.head.headersize=HEADER_SIZE_10;
	p.head.version=EIBNETIP_VERSION_10;
	p.head.servicetype=service_type;
	p.head.totalsize=HEADER_SIZE_10+data_size;
	p.data=senddata;
	
	// bug!!! p.data ist nur pointer -> das muss kontinuierlicher byte stream sein

	// prepare UDP header
	udph.udp_sport_h=EIBNETIP_PORT_NUMBER>>8;
	udph.udp_sport_l=EIBNETIP_PORT_NUMBER & 0x00FF;
	udph.udp_dport_h=receiver->port>>8;
	udph.udp_dport_l=receiver->port & 0x00FF;
	
	// prepare IP header
	iph.ip_daddr_3=receiver->ip>>24 & 0xFF;
	iph.ip_daddr_2=receiver->ip>>16 & 0xFF;
	iph.ip_daddr_1=receiver->ip>>8  & 0xFF;
	iph.ip_daddr_0=receiver->ip     & 0xFF;
	

	switch (receiver->hostprotocol) {
		case IPV4_UDP:
		
			// prepare IP header
			iph.ip_protocol=17;
		
			// send UDP packet
			udpsenddata((byte *)&p,p.head.totalsize,&udph,&iph);
			break;
			
		case IPV4_TCP:
			printfUART(UART_DEBUG,"eibnet/ip via tcp currently not supported...");
			break;
		default:
			printfUART(UART_DEBUG,"unsupported hostprotocol:0x%x",receiver->hostprotocol);
			break;	
	}
}

/**
 * helper function for preparing device information dib
 * memory has to be allocated outside
 **/
void prepare_device_information_dib(EIBNETIP_DEVINF_DIB *dib) {
	dib->structlength=sizeof(EIBNETIP_DEVINF_DIB);
	dib->descriptiontypecode=DEVICE_INFO;
	dib->knxmedium=CONFIG_KNXMEDIUM;
	dib->devicestatus=0x01; // program mode
	dib->eibaddress=KNX_INDIVIDUAL_ADDRESSES;
	dib->projectinstallationidentifier=PROJECT_INSTALLATION_ID;
	dib->serialnumber[0]=CONFIG_SN0;
	dib->serialnumber[1]=CONFIG_SN1;
	dib->serialnumber[2]=CONFIG_SN2;
	dib->serialnumber[3]=CONFIG_SN3;
	dib->serialnumber[4]=CONFIG_SN4;
	dib->serialnumber[5]=CONFIG_SN5;
	dib->multicastaddress[0]=SYSTEM_SETUP_MULTICAST_ADDRESS>>24 & 0xFF;
	dib->multicastaddress[1]=SYSTEM_SETUP_MULTICAST_ADDRESS>>16 & 0xFF;
	dib->multicastaddress[2]=SYSTEM_SETUP_MULTICAST_ADDRESS>>8  & 0xFF;
	dib->multicastaddress[3]=SYSTEM_SETUP_MULTICAST_ADDRESS     & 0xFF;
	dib->macaddress[0]=CS8900_MAC45>>8;
	dib->macaddress[1]=CS8900_MAC45;
	dib->macaddress[2]=CS8900_MAC23>>8;
	dib->macaddress[3]=CS8900_MAC23;
	dib->macaddress[4]=CS8900_MAC01>>8;
	dib->macaddress[5]=CS8900_MAC01;
	memcpy(dib->name,FRIENDLY_NAME,30);
}

/**
 * helper function for preparing supported services dib
 * memory has to be allocated outside
 **/
void prepare_supported_services_dib(EIBNETIP_SUPPFAM_DIB *supp,byte *supported) {
	int i=0;	// index to services data

	supp->structlength=2+(3*2); //length
	supp->descriptiontypecode=SUPP_SVC_FAMILIES;
	
	supported[i++]=EIBNETIP_CORE;
	supported[i++]=0x01;
	
	supported[i++]=EIBNETIP_DEVMGMT;
	supported[i++]=0x01;

	supported[i++]=EIBNETIP_TUNNELING;
	supported[i++]=0x01;

	supported[i++]=EIBNETIP_ROUTING;
	supported[i++]=0x01;

	
	supp->serviceidandversion=supported;
}

/**
 * handles search requests from eibnet/ip clients
 **/
void handleSearchRequest(byte * rcvdata, word rcvdatalen) {

	EIBNETIP_SEARCH_RESPONSE response;	// search response
	EIBNETIP_HPAI hpai_own; 				// own control endpoint
	EIBNETIP_DEVINF_DIB dib;				// device information dib
	EIBNETIP_SUPPFAM_DIB supp;				// supported services dib
	byte supported[3*2];		// supported services data
	EIBNETIP_MANUFACTURER_DIB manu;		// manufacturer data



	// extract discovery endpoint
	EIBNETIP_HPAI * hpai;
	hpai = (EIBNETIP_HPAI *)rcvdata[0];
	#ifdef DEBUG_ENABLE_LITTLE
	printfUART(UART_DEBUG,"received SearchRequest from %i.%i.%i.%i\r"
												,hpai->ip>>24 & 0xFF
												,hpai->ip>>16 & 0xFF
												,hpai->ip>>8  & 0xFF
												,hpai->ip     & 0xFF);
	#endif

	// prepare own hpai endpoint
	hpai_own.structlength=sizeof(EIBNETIP_HPAI); // struct_length
	hpai_own.hostprotocol=hpai->hostprotocol;
	hpai_own.ip=CURRENT_IP_ADDRESS;
	hpai_own.port=EIBNETIP_PORT_NUMBER;
	
	// prepare device information dib
	prepare_device_information_dib(&dib);

	// prepare supported services dib
	prepare_supported_services_dib(&supp, &supported[0]);

	// prepare manufacturer data dib
	// tbd...


	// prepare search response
	response.endpoint=&hpai_own;
	response.devicehardware=&dib;
	response.supported=&supp;
	response.manufacturer=NULL;
	
	eibnetipsend(hpai, SEARCH_RESPONSE, (byte *)&response, 
				sizeof(EIBNETIP_HPAI)+
				sizeof(EIBNETIP_DEVINF_DIB)+
				2+(3*2));
}


/**
 * handles description requests from clients
 **/
void handleDescriptionRequest(byte *rcvdata, word rcvdatalen) {

	EIBNETIP_DESCRIPTION_RESPONSE response;
	EIBNETIP_DEVINF_DIB dib;				// device information dib
	EIBNETIP_SUPPFAM_DIB supp;				// supported services dib
	byte supported[3*2];		// supported services data

	// extract control endpoint
	EIBNETIP_HPAI *hpai;
	hpai = (EIBNETIP_HPAI *)rcvdata[0];
	#ifdef DEBUG_ENABLE_LITTLE
	printfUART(UART_DEBUG,"received DescriptionRequest from %i.%i.%i.%i\r"
												,hpai->ip>>24 & 0xFF
												,hpai->ip>>16 & 0xFF
												,hpai->ip>>8  & 0xFF
												,hpai->ip     & 0xFF);
	#endif
	
	// prepare device information dib
	prepare_device_information_dib(&dib);

	// prepare supported services dib
	prepare_supported_services_dib(&supp, &supported[0]);

	// prepare description response
	response.devicehardware=&dib;
	response.supported=&supp;
	
	eibnetipsend(hpai, DESCRIPTION_RESPONSE, (byte *)&response, 
				sizeof(EIBNETIP_DEVINF_DIB)+
				2+(3*2));
	
}

/**
 * handles Connect Requests from client
 * and passes them to higher layers (tunneling, management,...)
 **/
void handleConnectRequest(byte *rcvdata, word rcvdatalen) {
	EIBNETIP_HPAI *hpai_control;
	EIBNETIP_HPAI *hpai_data;
	EIBNETIP_HPAI hpai_owndata={0};
	EIBNETIP_CRI_CRD *cri;
	EIBNETIP_CRI_CRD crd={0};
	EIBNETIP_CONNECT_RESPONSE response;
	byte error=E_NO_ERROR; // error code
	int i=0; // index variable
	int j=0; // index variable
	byte channelid=0;
	unsigned int *knxaddress=0; //knx individual address for tunneling c.
	

	// extract control endpoint
	hpai_control = (EIBNETIP_HPAI *)rcvdata[0];
	#ifdef DEBUG_ENABLE_LITTLE
	printfUART(UART_DEBUG,"received ConnectRequest from %i.%i.%i.%i\r"
												,hpai_control->ip>>24 & 0xFF
												,hpai_control->ip>>16 & 0xFF
												,hpai_control->ip>>8  & 0xFF
												,hpai_control->ip     & 0xFF);
	#endif

	// extract data endpoint
	hpai_data = (EIBNETIP_HPAI *)rcvdata[hpai_control->structlength];
	

	// extract cri
	cri = (EIBNETIP_CRI_CRD *)rcvdata[
						hpai_control->structlength
						+hpai_data->structlength];


	// prepare own hpai endpoint
	hpai_owndata.structlength=sizeof(EIBNETIP_HPAI); // struct_length
	hpai_owndata.hostprotocol=hpai_data->hostprotocol;
	hpai_owndata.ip=CURRENT_IP_ADDRESS;
	hpai_owndata.port=EIBNETIP_PORT_NUMBER;


	// check, if new connection is supported
	// check for E_NO_MORE_CONNECTIONS
	// if no error, assign channelid
	error=E_NO_MORE_CONNECTIONS;
	for (i=0;i<EIBNETIP_MAXCONNECTIONS;i++) {
		if (eibcon[i].channelid==0) {
			channelid=i+1; // channelid is always > 0 and <= EIBNET_MAXCONNECTIONS
			error=E_NO_ERROR;
			break;
		}
	}
	
	// check for E_CONNECTION_TYPE and E_CONNECTION_OPTION
	// if valid, pass to next layer
	switch (cri->connectiontypecode) {
		case (DEVICE_MGMT_CONNECTION):
			// prepare crd
			crd.structlength=0x02;
			crd.connectiontypecode=cri->connectiontypecode;
			crd.protocolindependentdata=NULL;
			crd.protocoldependentdata=NULL;
			error=E_NO_ERROR;
			break;

		case (TUNNEL_CONNECTION):
			if (cri->structlength == 0x04) {
				
				switch (cri->protocolindependentdata[0]) { // determine tunnelling layer
					case (TUNNEL_LINKLAYER):
						eibcon[channelid-1].connectioninfo = TUNNEL_LINKLAYER;
						
						#ifdef DEBUG_ENABLE_LITTLE
							printfUART(UART_DEBUG,"tunnelling connection in linklayer mode established\r");
						#endif
						break;
					
					case (TUNNEL_RAW):
						error=E_CONNECTION_TYPE;
						break;
						
					case (TUNNEL_BUSMONITOR):
						error=E_CONNECTION_TYPE;
						break;
						
					default:
						#ifdef DEBUG_ENABLE_LITTLE
							printfUART(UART_DEBUG,"unsupported tunnelling connection\r");
						#endif
						error=E_CONNECTION_TYPE;
						break;
				}
				
				if (error=E_NO_ERROR) {
					// determine knx address
					for (j=0;j<ADDITIONAL_INDIVIDUAL_ADDRESSES_NR;j++) {
						knxaddress=&ADDITIONAL_INDIVIDUAL_ADDRESSES[j];
	
						for (i=0;i<EIBNETIP_MAXCONNECTIONS;i++) {
							// check, if address is free
							if (eibcon[i].connectiontype==TUNNEL_CONNECTION && 
									knxaddress==eibcon[i].knxaddress) {
									// address not free
									knxaddress=NULL;
									break;
							}
						}
						
						if (knxaddress!=NULL) {
							if (cri->protocolindependentdata[0]==TUNNEL_LINKLAYER) {
								// generate iack messages
								addknxaddress(knxaddress);
							}
							break;
						}
					}
				}

				if (knxaddress==NULL)
					error=E_NO_MORE_CONNECTIONS;
			}
			else { // structure length not ok
				error=E_CONNECTION_TYPE;
			}

			// prepare crd
			crd.structlength=0x04;
			crd.connectiontypecode=cri->connectiontypecode;
			crd.protocolindependentdata=(byte *)knxaddress; // bug? word pointer
			crd.protocoldependentdata=NULL;

			break;

		case (REMLOG_CONNECTION):
			error=E_CONNECTION_TYPE;
			break;

		case (REMCONF_CONNECTION):
			error=E_CONNECTION_TYPE;
			break;

		case (OBJSVR_CONNECTION):
			error=E_CONNECTION_TYPE;
			break;

		default:
			#ifdef DEBUG_ENABLE_LITTLE
			printfUART(UART_DEBUG,"unsupported connection request. code:0x%e",cri->connectiontypecode);
			#endif
			break;
	}	


	// build response
	response.channelid=channelid;
	response.status=error;
	response.dataendpoint=&hpai_owndata;
	response.crd=&crd;

	// if no error occured, establish connection by adding parameters
	// to eibcon[channelid-1]
	if (error==E_NO_ERROR) {
		eibcon[channelid-1].channelid=channelid;
		eibcon[channelid-1].sequencecounter_rcv=-1;
		eibcon[channelid-1].sequencecounter_sent=0;
		eibcon[channelid-1].connectiontype=cri->connectiontypecode;
		eibcon[channelid-1].timeout=0;
		eibcon[channelid-1].status=E_NO_ERROR;
		eibcon[channelid-1].hpai=*hpai_data;
		eibcon[channelid-1].knxaddress=knxaddress;
	}
	
	// send response frame
	eibnetipsend(hpai_control, CONNECT_RESPONSE, (byte *)&response, 
				2+
				sizeof(EIBNETIP_HPAI)+
				crd.structlength);

}

/**
 * handles Connectionstate Requests from clients
 **/
void handleConnectionstateRequest(byte *rcvdata, word rcvdatalen) {
	EIBNETIP_HPAI *hpai_control;
	EIBNETIP_CONNECTION *c;
	byte response[2];

	response[1]=E_NO_ERROR;
	
	// extract channelid
	response[0]=rcvdata[0];
	
	// extract control endpoint
	hpai_control = (EIBNETIP_HPAI *)rcvdata[2];
	#ifdef DEBUG_ENABLE_LITTLE
	printfUART(UART_DEBUG,"received ConnectionstateRequest from %i.%i.%i.%i\r"
												,hpai_control->ip>>24 & 0xFF
												,hpai_control->ip>>16 & 0xFF
												,hpai_control->ip>>8  & 0xFF
												,hpai_control->ip     & 0xFF);
	#endif

	// get connectiion
	c=&eibcon[response[0]-1];

	// no active data connection with this channelid
	if (c->channelid==0) {
		response[1]=E_CONNECTION_ID;
	}
	else {
		// get status of data connection
		response[1]=c->status;
	}

	// send response frame
	eibnetipsend(hpai_control, CONNECTIONSTATE_RESPONSE, (byte *)&response, 2);
}

/**
 * handles Disconnect Requests from clients
 **/
void handleDisconnectRequest(byte *rcvdata, word rcvdatalen) {
	EIBNETIP_HPAI *hpai_control;
	EIBNETIP_CONNECTION *c;
	byte response[2];

	response[1]=E_NO_ERROR;
	
	// extract channelid
	response[0]=rcvdata[0];
	
	// extract control endpoint
	hpai_control = (EIBNETIP_HPAI *)rcvdata[2];
	#ifdef DEBUG_ENABLE_LITTLE
	printfUART(UART_DEBUG,"received ConnectionstateRequest from %i.%i.%i.%i\r"
												,hpai_control->ip>>24 & 0xFF
												,hpai_control->ip>>16 & 0xFF
												,hpai_control->ip>>8  & 0xFF
												,hpai_control->ip     & 0xFF);
	#endif

	// get connectiion
	c=&eibcon[response[0]-1];

	// close channel
	if (c->channelid!=0) {
		c->channelid=0;
		c->sequencecounter_rcv=0;
		c->sequencecounter_sent=0;
		c->connectiontype=0;
		c->timeout=0;
		c->status=E_NO_ERROR;
		if (c->knxaddress!=0) {
			remknxaddress(c->knxaddress);
		}
		memset(&c->hpai,0,sizeof(EIBNETIP_HPAI));
	}

	// send response frame
	eibnetipsend(hpai_control, DISCONNECT_RESPONSE, (byte *)&response, 2);
}
